GENOME_FILE_SUFFIX = "_genomic.fna"
PROTEIN_FILE_SUFFIX = "_protein.faa"
NT_GENE_FILE_SUFFIX = "_protein.fna"
GFF_FILE_SUFFIX = "_protein.gff"
TRANSLATION_TABLE_SUFFIX = "_translation_table.tsv"
CHECKSUM_SUFFIX = ".sha256"


TIGRFAM_SUFFIX = "_tigrfam.tsv"
TIGRFAM_TOP_HIT_SUFFIX = "_tigrfam_tophit.tsv"
PFAM_SUFFIX = "_pfam.tsv"
PFAM_TOP_HIT_SUFFIX = "_pfam_tophit.tsv"


# import prodigal
import math
import time
import random
import os

from itertools import islice
from biolib.seq_io import read_fasta
from biolib.common import remove_extension


##################################################
############MISC UTILITIES########################
##################################################

def add_ncbi_prefix(refname):
    if refname.startswith("GCF_"):
        return "RS_" + refname
    elif refname.startswith("GCA_"):
        return "GB_" + refname
    else:
        return refname


def splitchunks(d, n):
    chunksize = int(math.ceil(len(d) / float(n)))
    it = iter(d)
    for _ in xrange(0, len(d), chunksize):
        yield {k: d[k] for k in islice(it, chunksize)}


def splitchunks_list(l, n):
    """Yield successive n-sized chunks from l."""
    chunksize = int(math.ceil(len(l) / float(n)))
    for i in range(0, len(l), chunksize):
        yield l[i:i + chunksize]


def generateTempTableName():
    rng = random.SystemRandom()
    suffix = ''
    for _ in range(0, 10):
        suffix += rng.choice(
            'abcefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ')
    return "TEMP" + suffix + str(int(time.time()))


def list_genomes_dir(userdir):
    """List fasta files in a specified directory

    Parameters
    ----------
    userdir : str
        Directory path where all fasta files are

    Returns
    -------
    dict
        Dictionary indicating the genomic file for each genome.
    """
    if not os.path.exists(userdir):
        raise ValueError('{0} does not exist.'.format(userdir))
    else:
        onlygenomefiles = {f: os.path.join(userdir, f) for f in os.listdir(
            userdir) if os.path.isfile(os.path.join(userdir, f))}
        for potential_file in onlygenomefiles:
            try:
                read_fasta(os.path.join(userdir, potential_file))
            except:
                raise IOError("{0} is not a fasta file." .format(
                    os.path.join(userdir, potential_file)))
        return onlygenomefiles


def merge_two_dicts(x, y):
    '''Given two dicts, merge them into a new dict as a shallow copy.'''
    z = x.copy()
    z.update(y)
    return z



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging
import shutil
import multiprocessing as mp
from collections import defaultdict

from biolib.external.prodigal import (Prodigal as BioLibProdigal)


class Prodigal(object):
    """Perform ab initio gene prediction using Prodigal."""

    def __init__(self,
                    threads,
                    proteins,
                    marker_gene_dir, 
                    protein_file_suffix, 
                    nt_gene_file_suffix, 
                    gff_file_suffix):
        """Initialize."""

        self.logger = logging.getLogger('timestamp')

        self.threads = threads
        
        self.proteins = proteins

        self.marker_gene_dir = marker_gene_dir
        self.protein_file_suffix = protein_file_suffix
        self.nt_gene_file_suffix = nt_gene_file_suffix
        self.gff_file_suffix = gff_file_suffix

    def _run_prodigal(self, genome_id, fasta_path):
        """Run Prodigal.

        Parameters
        ----------
        fasta_path : str
            Path to FASTA file to process.
        """

        output_dir = os.path.join(self.marker_gene_dir, genome_id)

        prodigal = BioLibProdigal(1, False)
        summary_stats = prodigal.run([fasta_path], output_dir, called_genes=self.proteins)
        summary_stats = summary_stats[summary_stats.keys()[0]]

        # rename output files to adhere to GTDB conventions and desired genome ID
        aa_gene_file = os.path.join(output_dir, genome_id + self.protein_file_suffix)
        shutil.move(summary_stats.aa_gene_file, aa_gene_file)

        nt_gene_file = None
        gff_file = None
        translation_table_file = None
        if not self.proteins:
            nt_gene_file = os.path.join(output_dir, genome_id + self.nt_gene_file_suffix)
            shutil.move(summary_stats.nt_gene_file, nt_gene_file)

            gff_file = os.path.join(output_dir, genome_id + self.gff_file_suffix)
            shutil.move(summary_stats.gff_file, gff_file)

            # save translation table information
            translation_table_file = os.path.join(output_dir, 'prodigal_translation_table.tsv')
            fout = open(translation_table_file, 'w')
            fout.write('%s\t%d\n' % ('best_translation_table', summary_stats.best_translation_table))
            fout.write('%s\t%.2f\n' % ('coding_density_4', summary_stats.coding_density_4 * 100))
            fout.write('%s\t%.2f\n' % ('coding_density_11', summary_stats.coding_density_11 * 100))
            fout.close()

        return (aa_gene_file, nt_gene_file, gff_file, translation_table_file)

    def _worker(self, out_dict, worker_queue, writer_queue):
        """This worker function is invoked in a process."""

        while True:
            data = worker_queue.get(block=True, timeout=None)
            if data is None:
                break

            genome_id, file_path = data

            rtn_files = self._run_prodigal(genome_id, file_path)
            aa_gene_file, nt_gene_file, gff_file, translation_table_file = rtn_files
            file_paths = {}
            file_paths["aa_gene_path"] = aa_gene_file
            file_paths["nt_gene_path"] = nt_gene_file
            file_paths["gff_path"] = gff_file
            file_paths["translation_table_path"] = translation_table_file

            out_dict[genome_id] = file_paths
            writer_queue.put(genome_id)

    def _writer(self, num_items, writer_queue):
        """Store or write results of worker threads in a single thread."""
        processed_items = 0
        while processed_items < num_items:
            a = writer_queue.get(block=True, timeout=None)
            if a is None:
                break

            processed_items += 1
            statusStr = '==> Finished processing %d of %d (%.1f%%) genomes.' % (processed_items,
                                                                                num_items,
                                                                                float(processed_items) * 100 / num_items)
            sys.stdout.write('%s\r' % statusStr)
            sys.stdout.flush()

        sys.stdout.write('\n')

    def run(self, genomic_files):
        """Run Prodigal across a set of genomes.

        Parameters
        ----------
        genomic_files : dict
            Dictionary indicating the genomic and gene file for each genome.
        """
        
        # populate worker queue with data to process
        worker_queue = mp.Queue()
        writer_queue = mp.Queue()

        for genome_id, file_path in genomic_files.iteritems():
            worker_queue.put([genome_id, file_path])

        for _ in range(self.threads):
            worker_queue.put(None)

        try:
            manager = mp.Manager()
            out_dict = manager.dict()

            worker_proc = [mp.Process(target=self._worker, args=(out_dict, 
                                                                    worker_queue, 
                                                                    writer_queue)) for _ in range(self.threads)]
            writer_proc = mp.Process(target=self._writer, args=(len(genomic_files), writer_queue))

            writer_proc.start()
            for p in worker_proc:
                p.start()

            for p in worker_proc:
                p.join()

            writer_queue.put(None)
            writer_proc.join()
        except:
            for p in worker_proc:
                p.terminate()

            writer_proc.terminate()
            raise
            
        result_dict = {k: v for k, v in out_dict.items()}
        return result_dict



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import sys
import random
import logging

import dendropy


class RerootTree(object):
    """Reroot tree."""

    def __init__(self):
        """Initialization."""
        self.logger = logging.getLogger('timestamp')

    def root_with_outgroup(self, input_tree, output_tree, outgroup):
        """Reroot the tree using the given outgroup.

        Parameters
        ----------
        input_tree : str
          File containing Newick tree to rerooted.
        output_tree : str
          Name of file for rerooted tree.
        outgroup : iterable
          Labels of taxa in outgroup.
        """

        tree = dendropy.Tree.get_from_path(input_tree,
                                           schema='newick',
                                           rooting='force-rooted',
                                           preserve_underscores=True)

        outgroup = set(outgroup)
        outgroup_in_tree = set()
        ingroup_leaves = set()
        for n in tree.leaf_node_iter():
            if n.taxon.label in outgroup:
                outgroup_in_tree.add(n.taxon)
            else:
                ingroup_leaves.add(n)

        self.logger.info('Identified %d outgroup taxa in the tree.' %
                         len(outgroup_in_tree))
        self.logger.info('Identified %d ingroup taxa in the tree.' %
                         len(ingroup_leaves))

        if len(outgroup_in_tree) == 0:
            self.logger.warning('No outgroup taxa identified in the tree.')
            self.logger.warning('Tree was not rerooted.')
            sys.exit(0)

        # Since finding the MRCA is a rooted tree operation,
        # the tree is first rerooted on an ingroup taxa. This
        # ensures the MRCA of the outgroup can be identified
        # so long as the outgroup is monophyletic. If the
        # outgroup is polyphyletic trying to root on it
        # is ill defined. To try and pick a "good" root for
        # polyphyletic outgroups, random ingroup taxa are
        # selected until two of them give the same size
        # lineage. This will, likely, be the smallest
        # bipartition possible for the given outgroup though
        # this is not guaranteed.
        mrca = tree.mrca(taxa=outgroup_in_tree)
        mrca_leaves = len(mrca.leaf_nodes())
        while True:
            rnd_ingroup = random.sample(ingroup_leaves, 1)[0]
            tree.reroot_at_edge(rnd_ingroup.edge,
                                length1=0.5 * rnd_ingroup.edge_length,
                                length2=0.5 * rnd_ingroup.edge_length)

            mrca = tree.mrca(taxa=outgroup_in_tree)
            if len(mrca.leaf_nodes()) == mrca_leaves:
                break

            mrca_leaves = len(mrca.leaf_nodes())

        if len(mrca.leaf_nodes()) != len(outgroup_in_tree):
            self.logger.info(
                'Outgroup is not monophyletic. Tree will be rerooted at the MRCA of the outgroup.')
            self.logger.info('The outgroup consisted of %d taxa, while the MRCA has %d leaf nodes.' % (
                len(outgroup_in_tree), len(mrca.leaf_nodes())))
            if len(mrca.leaf_nodes()) == len(tree.leaf_nodes()):
                self.logger.warning('The MRCA spans all taxa in the tree.')
                self.logger.warning(
                    'This indicating the selected outgroup is likely polyphyletic in the current tree.')
                self.logger.warning(
                    'Polyphyletic outgroups are not suitable for rooting. Try another outgroup.')
        else:
            self.logger.info('Outgroup is monophyletic.')

        if mrca.edge_length is None:
            self.logger.info(
                'Tree appears to already be rooted on this outgroup.')
        else:
            self.logger.info('Rerooting tree.')
            tree.reroot_at_edge(mrca.edge,
                                length1=0.5 * mrca.edge_length,
                                length2=0.5 * mrca.edge_length)
            tree.write_to_path(output_tree, schema='newick',
                               suppress_rooting=True, unquoted_underscores=True)
            self.logger.info('Rerooted tree written to: %s' % output_tree)

    def midpoint(self, input_tree, output_tree):
        """Reroot tree bat midpoint.

        Parameters
        ----------
        input_tree : str
          File containing Newick tree to rerooted.
        output_tree : str
          Name of file for rerooted tree.
        """

        tree = dendropy.Tree.get_from_path(
            input_tree, schema='newick', rooting='force-rooted', preserve_underscores=True)
        tree.reroot_at_midpoint()
        tree.write_to_path(output_tree, schema='newick',
                           suppress_rooting=True, unquoted_underscores=True)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import shutil
import logging
import tempfile
import random
import multiprocessing
import config.config as Config
import dendropy
import cmd

from collections import defaultdict
from biolib.common import remove_extension, make_sure_path_exists
from biolib.seq_io import read_seq, read_fasta
from biolib.newick import parse_label
from biolib.external.execute import check_dependencies
from biolib.taxonomy import Taxonomy
from numpy import median as np_median
from tools import add_ncbi_prefix, splitchunks
from relative_distance import RelativeDistance
from operator import itemgetter

sys.setrecursionlimit(15000)


class Classify():
    """Determine taxonomic classification of genomes by ML placement."""

    def __init__(self, cpus=1):
        """Initialize."""

        check_dependencies(['pplacer', 'guppy', 'fastANI'])

        self.taxonomy_file = Config.TAXONOMY_FILE
        self.gtdb_taxonomy = Taxonomy().read(self.taxonomy_file)

        self.order_rank = ["d__", "p__", "c__", "o__", 'f__', 'g__', 's__']

        self.logger = logging.getLogger('timestamp')
        self.cpus = cpus

    def place_genomes(self,
                      user_msa_file,
                      marker_set_id,
                      out_dir,
                      prefix):
        """Place genomes into reference tree using pplacer."""
        # rename user MSA file for compatibility with pplacer
        if not user_msa_file.endswith('.fasta'):
            t = os.path.join(out_dir, prefix + '.user_msa.fasta')
            shutil.copyfile(user_msa_file, t)
            user_msa_file = t

        # run pplacer to place bins in reference genome tree
        num_genomes = sum([1 for _seq_id, _seq in read_seq(user_msa_file)])

        # get path to pplacer reference package
        if marker_set_id == 'bac120':
            self.logger.info(
                'Placing {} bacterial genomes into reference tree with pplacer (be patient).'.format(num_genomes))
            pplacer_ref_pkg = os.path.join(
                Config.PPLACER_DIR, Config.PPLACER_BAC120_REF_PKG)
        elif marker_set_id == 'ar122':
            self.logger.info(
                'Placing {} archaeal genomes into reference tree with pplacer (be patient).'.format(num_genomes))
            pplacer_ref_pkg = os.path.join(
                Config.PPLACER_DIR, Config.PPLACER_AR122_REF_PKG)
        elif marker_set_id == 'rps23':
            self.logger.info(
                'Placing {} genomes into reference tree with pplacer (be patient).'.format(num_genomes))
            pplacer_ref_pkg = os.path.join(
                Config.PPLACER_DIR, Config.PPLACER_RPS23_REF_PKG)

        # create pplacer output directory
        pplacer_out_dir = os.path.join(out_dir, 'pplacer')
        if not os.path.exists(pplacer_out_dir):
            os.makedirs(pplacer_out_dir)

        # run pplacer
        pplacer_out = os.path.join(
            pplacer_out_dir, 'pplacer.{}.out'.format(marker_set_id))
        pplacer_json_out = os.path.join(
            pplacer_out_dir, 'pplacer.{}.json'.format(marker_set_id))
        cmd = 'pplacer -j {} -c {} -o {} {} > {}'.format(self.cpus,
                                                         pplacer_ref_pkg,
                                                         pplacer_json_out,
                                                         user_msa_file,
                                                         pplacer_out)
        os.system(cmd)

        # extract tree
        tree_file = os.path.join(
            out_dir, prefix + ".{}.classify.tree".format(marker_set_id))
        cmd = 'guppy tog -o {} {}'.format(tree_file, pplacer_json_out)
        os.system(cmd)

        return tree_file

    def standardise_taxonomy(self, taxstring, marker_set=None):
        """Create a 7 rank taxonomy string from an incomplete taxonomy string

        Parameters
        ----------
        tax_string : incomplete taxonomy string

        Returns
        -------
        string
            7 rank taxonomy string.
        """
        # return taxstring
        taxlist = taxstring.split(";")
        if marker_set == 'bac120':
            taxlist.insert(0, 'd__Bacteria')
        if marker_set == 'ar122':
            taxlist.insert(0, 'd__Archaea')
        taxlist.extend(self.order_rank[len(taxlist):])
        new_taxstring = ";".join(taxlist)
        return new_taxstring

    def _write_red_dict(self, out_dir, prefix, marker_set_id):
        """Write the RED value for each rank to a file

        Parameters
        ----------
        out_dir : output directory
        prefix : desired prefix for output files
        marker_set_id : bacterial or archeal id (bac120 or ar122)

        Returns
        -------
        dictionary
            dictionary[rank_prefix] = red_value

        """

        reddictfile = open(os.path.join(
            out_dir, prefix + '.{}.red_dictionary.tsv'.format(marker_set_id)), 'w')

        marker_dict = {}
        if marker_set_id == 'bac120':
            marker_dict = Config.RED_DIST_BAC_DICT
        elif marker_set_id == 'ar122':
            marker_dict = Config.RED_DIST_ARC_DICT
        reddictfile.write('Phylum\t{}\n'.format(marker_dict.get('p__')))
        reddictfile.write('Class\t{}\n'.format(marker_dict.get('c__')))
        reddictfile.write('Order\t{}\n'.format(marker_dict.get('o__')))
        reddictfile.write('Family\t{}\n'.format(marker_dict.get('f__')))
        reddictfile.write('Genus\t{}\n'.format(marker_dict.get('g__')))
        reddictfile.close()
        return marker_dict

    def run(self,
            genomes,
            align_dir,
            out_dir,
            prefix,
            debugopt=False):
        try:
            """Classify genomes based on position in reference tree."""

            for marker_set_id in ('bac120', 'ar122'):
                user_msa_file = os.path.join(
                    align_dir, prefix + '.{}.user_msa.fasta'.format(marker_set_id))
                if not os.path.exists(user_msa_file):
                    # file will not exist if there are no User genomes from a
                    # given domain
                    continue

                classify_tree = self.place_genomes(user_msa_file,
                                                   marker_set_id,
                                                   out_dir,
                                                   prefix)

                # get taxonomic classification of each user genome
                tree = dendropy.Tree.get_from_path(classify_tree,
                                                   schema='newick',
                                                   rooting='force-rooted',
                                                   preserve_underscores=True)

                summaryfout = open(os.path.join(
                    out_dir, prefix + '.{}.summary.tsv'.format(marker_set_id)), 'w')
                if debugopt:
                    debugfile = open(os.path.join(
                        out_dir, prefix + '.{}.debug_file.tsv'.format(marker_set_id)), 'w')

                marker_dict = self._write_red_dict(
                    out_dir, prefix, marker_set_id)

                summaryfout.write("user_genome\tclassification\tfastani_reference\tfastani_taxonomy\tfastani_ani\tfastani_af\t" +
                                  "closest_placement_reference\tclosest_placement_taxonomy\tclosest_placement_ani\tclosest_placement_af\t" +
                                  "classification_method\tnote\tother_related_references(genome_id,species_name,ANI,AF)\n")
                if debugopt:
                    debugfile.write(
                        "User genome\tRed value\tHigher rank\tHigher value\tLower rank\tLower value\tcase\tclosest_rank\ttool\n")

                # Genomes can be classified by using FastANI or RED values
                # We go through all leaves of the tree. if the leaf is a user
                # genome we take it's parent node and look at all the leaves
                # for this node.
                all_fastani_dict = {}
                self.logger.info(
                    'Calculating Average Nucleotide Identity using FastANI.')
                fastani_verification = {}
                number_comparison = 0
                for userleaf in tree.leaf_node_iter():
                    # for each user genome, we select the first parent node with a label.
                    # if, while going up the tree, we find a node with only one
                    # reference genome, we select this reference genome as
                    # leaf_reference.
                    if userleaf.taxon.label[0:3] not in ['RS_', 'GB_', 'UBA']:

                        par_node = userleaf.parent_node
                        leaf_ref_genome = None
                        leaf_ref_genomes = [subnd for subnd in par_node.leaf_iter(
                        ) if subnd.taxon.label.replace("'", '')[0:3] in ['RS_', 'GB_', 'UBA']]
                        if len(leaf_ref_genomes) == 1:
                            leaf_ref_genome = leaf_ref_genomes[0]

                        _support, parent_taxon, _aux_info = parse_label(
                            par_node.label)
                        # while par_node is not None and
                        # (par_node.distance_from_root() >
                        # marker_dict.get('f__') or (parent_taxon is not None
                        # and parent_taxon.split(";")[-1].startswith('g__'))):
                        while par_node is not None and not parent_taxon:
                            par_node = par_node.parent_node
                            if leaf_ref_genome is None:
                                leaf_ref_genomes = [subnd for subnd in par_node.leaf_iter(
                                ) if subnd.taxon.label.replace("'", '')[0:3] in ['RS_', 'GB_', 'UBA']]
                                if len(leaf_ref_genomes) == 1:
                                    leaf_ref_genome = leaf_ref_genomes[0]
                            _support, parent_taxon, _aux_info = parse_label(
                                par_node.label)

                        # if the parent node is at the genus level
                        parent_rank = parent_taxon.split(";")[-1]
                        if parent_rank.startswith('g__'):
                            # we get all the reference genomes under this genus
                            list_subnode_initials = [subnd.taxon.label.replace(
                                "'", '')[0:3] for subnd in par_node.leaf_iter()]
                            if (list_subnode_initials.count('RS_') + list_subnode_initials.count('GB_') + list_subnode_initials.count('UBA')) < 1:
                                raise Exception(
                                    "There is no reference genomes under '{}'".format('parent_rank'))
                            else:
                                dict_dist_refgenomes = {}
                                list_ref_genomes = [subnd for subnd in par_node.leaf_iter(
                                ) if subnd.taxon.label.replace("'", '')[0:3] in ['RS_', 'GB_', 'UBA']]
                                # we pick the first 100 genomes closest to the
                                # user genome under the same genus
                                for ref_genome in list_ref_genomes:
                                    taxon_labels = [
                                        userleaf.taxon.label, ref_genome.taxon.label]
                                    mrca = tree.mrca(taxon_labels=taxon_labels)
                                    # the following command is faster than
                                    # calculating the patristic distance
                                    dict_dist_refgenomes[ref_genome] = (userleaf.distance_from_root(
                                    ) - mrca.distance_from_root()) + (ref_genome.distance_from_root() - mrca.distance_from_root())
                                sorted_l = sorted(
                                    dict_dist_refgenomes.iteritems(), key=itemgetter(1))
                                sorted_l = sorted_l[0:100]
                                number_comparison += len(sorted_l)
                                fastani_verification[userleaf] = {
                                    "potential_g": sorted_l, "pplacer_g": leaf_ref_genome}
                        else:
                            if leaf_ref_genome:
                                fastani_verification[userleaf] = {"potential_g": [
                                    (leaf_ref_genome, 0.0)], "pplacer_g": leaf_ref_genome}

                # self.logger.info('{} need to be compared.'.format(number_comparison))

                manager = multiprocessing.Manager()
                out_q = manager.dict()
                procs = []
                nprocs = self.cpus

                if len(fastani_verification) > 0:
                    for item in splitchunks(fastani_verification, nprocs):
                        p = multiprocessing.Process(
                            target=self._fastaniWorker,
                            args=(item, genomes, out_q))
                        procs.append(p)
                        p.start()

                    # Wait for all worker processes to finish
                    for p in procs:
                        p.join()
                        if p.exitcode == 1:
                            raise ValueError("Stop!!")

                    all_fastani_dict = dict(out_q)

                classified_user_genomes = self._sort_fastani_results(
                    fastani_verification, all_fastani_dict, summaryfout)

                self.logger.info('{0} genomes have been classify using FastANI and Pplacer.'.format(
                    len(classified_user_genomes)))

                scaled_tree = self._calculate_red_distances(
                    classify_tree, out_dir)

                user_genome_ids = set(read_fasta(user_msa_file).keys())
                # we remove ids aleady classified with FastANI
                user_genome_ids = user_genome_ids.difference(
                    set(classified_user_genomes))
                for leaf in scaled_tree.leaf_node_iter():
                    if leaf.taxon.label in user_genome_ids:
                        # In some cases , pplacer can associate 2 user genomes
                        # on the same parent node so we need to go up the tree
                        # to find a node with a reference genome as leaf.
                        cur_node = leaf.parent_node
                        list_subnode_initials = [subnd.taxon.label.replace(
                            "'", '')[0:3] for subnd in cur_node.leaf_iter()]
                        while 'RS_' not in list_subnode_initials and 'GB_' not in list_subnode_initials and 'UBA' not in list_subnode_initials:
                            cur_node = cur_node.parent_node
                            list_subnode_initials = [subnd.taxon.label.replace(
                                "'", '')[0:3] for subnd in cur_node.leaf_iter()]

                        current_rel_list = cur_node.rel_dist

                        parent_taxon_node = cur_node.parent_node
                        _support, parent_taxon, _aux_info = parse_label(
                            parent_taxon_node.label)

                        while parent_taxon_node is not None and not parent_taxon:
                            parent_taxon_node = parent_taxon_node.parent_node
                            _support, parent_taxon, _aux_info = parse_label(
                                parent_taxon_node.label)

                        # is the node represent multiple ranks, we select the lowest one
                        # i.e. if node is p__A;c__B;o__C we pick o__
                        parent_rank = parent_taxon.split(";")[-1][0:3]
                        parent_rel_dist = parent_taxon_node.rel_dist

                        debug_info = [leaf.taxon.label, parent_rank,
                                      parent_rel_dist, '', '', '', '']

                        child_taxons = []
                        closest_rank = None
                        detection = "taxonomic novelty determined using RED"
                        # if the genome is not placed between the genus and
                        # specie ranks
                        if parent_rank != 'g__':
                            # we select the child rank (if parent_rank = 'c__'
                            # child rank will be 'o__)'
                            child_rk = self.order_rank[self.order_rank.index(
                                parent_rank) + 1]

                            # get all reference genomes under the current node
                            list_subnode = [childnd.taxon.label.replace("'", '') for childnd in cur_node.leaf_iter(
                            ) if childnd.taxon.label[0:3] in ['RS_', 'UBA', 'GB_']]

                            # get all names for the child rank
                            list_ranks = [self.gtdb_taxonomy.get(
                                name)[self.order_rank.index(child_rk)] for name in list_subnode]

                            # if there is just one rank name
                            if len(set(list_ranks)) == 1:
                                for subranknd in cur_node.preorder_iter():
                                    _support, subranknd_taxon, _aux_info = parse_label(
                                        subranknd.label)
                                    if subranknd.is_internal() and subranknd_taxon is not None and subranknd_taxon.startswith(child_rk):
                                        child_taxons = subranknd_taxon.split(
                                            ";")
                                        child_taxon_node = subranknd
                                        child_rel_dist = child_taxon_node.rel_dist
                                        break
                            else:
                                # case 2a and 2b
                                closest_rank = parent_rank
                                detection = "taxonomic classification fully defined by topology"
                        else:
                            # case 1a
                            closest_rank = parent_rank
                            detection = "taxonomic classification fully defined by topology"

                        # case 1b
                        if len(child_taxons) == 0 and closest_rank is None:
                            list_leaves = [childnd.taxon.label.replace("'", '') for childnd in cur_node.leaf_iter(
                            ) if childnd.taxon.label[0:3] in ['RS_', 'UBA', 'GB_']]
                            if len(list_leaves) != 1:
                                raise Exception(
                                    'There should be only one leaf.')
                                sys.exit(-1)
                            list_leaf_ranks = self.gtdb_taxonomy.get(
                                list_leaves[0])[self.order_rank.index(child_rk):-1]
                            for leaf_taxon in reversed(list_leaf_ranks):
                                if leaf_taxon == list_leaf_ranks[0]:
                                    if abs(current_rel_list - marker_dict.get(leaf_taxon[:3])) < abs((current_rel_list) - marker_dict.get(parent_rank)):
                                        # and current_rel_list -
                                        # marker_dict.get(leaf_taxon[:3]) > 0
                                        # ):
                                        closest_rank = leaf_taxon[:3]
                                        debug_info[3] = leaf_taxon
                                        debug_info[5] = 'case 1b - III'
                                        break
                                else:
                                    pchildrank = list_leaf_ranks[list_leaf_ranks.index(
                                        leaf_taxon) - 1]
                                    if abs(current_rel_list - marker_dict.get(leaf_taxon[:3])) < abs(current_rel_list - marker_dict.get(pchildrank[:3])):
                                        closest_rank = leaf_taxon[:3]
                                        debug_info[1] = pchildrank
                                        debug_info[2] = 1.0
                                        debug_info[3] = leaf_taxon
                                        debug_info[5] = 'case 1b - II'
                                        break
                            if closest_rank is None:
                                closest_rank = parent_rank
                                debug_info[3] = list_leaf_ranks[0]
                                debug_info[5] = 'case 1b - IV'

                        # if there is multiple ranks on the child node (i.e genome between p__Nitrospirae and c__Nitrospiria;o__Nitrospirales;f__Nitropiraceae)
                        # we loop through the list of rank from f_ to c_ rank
                        for child_taxon in reversed(child_taxons):
                            # if lower rank is c__Nitropiria
                            if child_taxon == child_taxons[0]:
                                if (abs(current_rel_list - marker_dict.get(child_taxon[:3])) < abs(child_rel_dist - marker_dict.get(child_taxon[:3])) and
                                        abs(current_rel_list - marker_dict.get(child_taxon[:3])) < abs(current_rel_list - marker_dict.get(parent_rank))):
                                    debug_info[3] = ';'.join(child_taxons)
                                    debug_info[4] = child_rel_dist
                                    debug_info[5] = 'case 3b - II'
                                    closest_rank = child_taxon[:3]
                                elif closest_rank is None:
                                    closest_rank = parent_rank
                                    debug_info[3] = ';'.join(child_taxons)
                                    debug_info[4] = child_rel_dist
                                    debug_info[5] = 'case 3b - III'
                            else:
                                pchildrank = child_taxons[child_taxons.index(
                                    child_taxon) - 1]
                                if (abs(current_rel_list - marker_dict.get(child_taxon[:3])) < abs(current_rel_list - marker_dict.get(pchildrank[:3])) and
                                        abs(current_rel_list - marker_dict.get(child_taxon[:3])) < abs(child_rel_dist - marker_dict.get(child_taxon[:3]))):
                                    closest_rank = child_taxon
                                    debug_info[3] = ';'.join(child_taxons)
                                    debug_info[4] = child_rel_dist
                                    debug_info[5] = 'case 3b - I'
                                    break

                        # case 1b
                        if closest_rank is None:
                            raise Exception('closest rank is None')

                        debug_info[6] = closest_rank

                        list_subnode = [subnd.taxon.label.replace(
                            "'", '') for subnd in cur_node.leaf_iter()]
                        red_taxonomy = self._get_redtax(
                            list_subnode, closest_rank)

                        del debug_info[0]

                        summary_list = [None] * 13
                        summary_list[0] = leaf.taxon.label
                        summary_list[1] = self.standardise_taxonomy(
                            red_taxonomy)
                        summary_list[10] = detection
                        summaryfout.write("{0}\n".format(
                            '\t'.join(['None' if x is None else str(x) for x in summary_list])))
                        if debugopt:
                            debugfile.write('{0}\t{1}\t{2}\t{3}\n'.format(
                                leaf.taxon.label, current_rel_list, '\t'.join(str(x) for x in debug_info), detection))

                summaryfout.close()
                if debugopt:
                    debugfile.close()

                self._get_pplacer_taxonomy(
                    out_dir, prefix, marker_set_id, user_msa_file, tree)

        except ValueError as error:
            print "GTDB-Tk has stopped before finishing"
            print error
            raise
        except Exception as error:
            print "GTDB-Tk has stopped before finishing"
            print error
            raise

    def _remove_named_nodes(self, node):
        if node.is_internal():
            for childnode in node.child_nodes():
                _support, taxon, _aux_info = parse_label(childnode.label)
                if childnode.is_internal() and taxon:
                    node.remove_child(childnode)
                elif childnode.is_internal():
                    new_leaf_node = self._remove_named_nodes(childnode)
                    if new_leaf_node:
                        node.remove_child(childnode)
            if node.is_leaf():
                return node
            else:
                return False

    def _get_pplacer_taxonomy(self, out_dir, prefix, marker_set_id, user_msa_file, tree):
        """Parse the pplacer tree and write the partial taxonomy for each user genome based on their placements

        Parameters
        ----------
        out_dir : output directory
        prefix : desired prefix for output files
        marker_set_id : bacterial or archeal id (bac120 or ar122)
        user_msa_file : msa file listing all user genomes for a certain domain
        tree : pplacer tree including the user genomes

        Returns
        -------
        True

        """
        pplaceout = open(os.path.join(
            out_dir, prefix + '.{}.classification_pplacer.tsv'.format(marker_set_id)), 'w')

        # We get the pplacer taxonomy for comparison

        user_genome_ids = set(read_fasta(user_msa_file).keys())
        for leaf in tree.leaf_node_iter():
            if leaf.taxon.label in user_genome_ids:
                taxa = []
                cur_node = leaf
                while cur_node.parent_node:
                    _support, taxon, _aux_info = parse_label(cur_node.label)
                    if taxon:
                        for t in taxon.split(';')[::-1]:
                            taxa.append(t.strip())
                    cur_node = cur_node.parent_node
                taxa_str = ';'.join(taxa[::-1])
                pplaceout.write('{}\t{}\n'.format(
                    leaf.taxon.label, self.standardise_taxonomy(taxa_str, marker_set_id)))
        pplaceout.close()
        return True

    def _formatnote(self, sorted_dict, labels):
        """Format the note field by concatenating all information in a sorted dictionary

        Parameters
        ----------
        sorted_dict : sorted dictionary listing reference genomes, ani and alignement fraction for a specific user genome
                    (genomeid, {ani: value, af: value})
        labels : array of label that are removed from the note field

        Returns
        -------
        string
            note field

        """
        note_list = []
        for element in sorted_dict:
            if element[0] not in labels:
                note_str = "{}, {}, {}, {}".format(element[0], self.gtdb_taxonomy.get(
                    add_ncbi_prefix(element[0]))[6], element[1].get('ani'), element[1].get('af'))
                note_list.append(note_str)
        return note_list

    def _sort_fastani_results(self, fastani_verification, all_fastani_dict, summaryfout):
        classified_user_genomes = []
        for userleaf, potential_nodes in fastani_verification.iteritems():
            summary_list = [None] * 13
            if potential_nodes.get("pplacer_g"):
                pplacer_leafnode = potential_nodes.get("pplacer_g").taxon.label
                if pplacer_leafnode[0:3] in ['RS_', 'GB_']:
                    pplacer_leafnode = pplacer_leafnode[3:]
                if userleaf.taxon.label in all_fastani_dict:
                    sorted_dict = sorted(all_fastani_dict.get(
                        userleaf.taxon.label).iteritems(), key=lambda(_x, y): y['ani'], reverse=True)
                    fastani_matching_reference = sorted_dict[0][0]
                    taxa_str = ";".join(self.gtdb_taxonomy.get(
                        add_ncbi_prefix(pplacer_leafnode)))

                    summary_list[0] = userleaf.taxon.label
                    summary_list[2] = fastani_matching_reference
                    summary_list[3] = ";".join(self.gtdb_taxonomy.get(
                        add_ncbi_prefix(fastani_matching_reference)))
                    summary_list[4] = all_fastani_dict.get(userleaf.taxon.label).get(
                        fastani_matching_reference).get('ani')
                    summary_list[5] = all_fastani_dict.get(userleaf.taxon.label).get(
                        fastani_matching_reference).get('af')
                    summary_list[10] = 'ANI/Placement'

                    if pplacer_leafnode == fastani_matching_reference:
                        if taxa_str.endswith("s__"):
                            taxa_str = taxa_str + pplacer_leafnode
                        summary_list[1] = self.standardise_taxonomy(taxa_str)
                        summary_list[6] = summary_list[2]
                        summary_list[7] = summary_list[3]
                        summary_list[8] = summary_list[4]
                        summary_list[9] = summary_list[5]
                        summary_list[11] = 'topological placement and ANI have congruent species assignments'
                        if len(sorted_dict) > 0:
                            summary_list[12] = '; '.join(self._formatnote(
                                sorted_dict, [fastani_matching_reference]))

                        summaryfout.write("{}\n".format(
                            '\t'.join(['N/A' if x is None else str(x) for x in summary_list])))

                    else:
                        taxa_str = taxa_str[0:taxa_str.rindex(';')]
                        summary_list[1] = self.standardise_taxonomy(taxa_str)
                        summary_list[6] = pplacer_leafnode
                        summary_list[7] = ";".join(self.gtdb_taxonomy.get(
                            add_ncbi_prefix(pplacer_leafnode)))
                        if pplacer_leafnode in all_fastani_dict.get(userleaf.taxon.label):
                            summary_list[8] = all_fastani_dict.get(
                                userleaf.taxon.label).get(pplacer_leafnode).get('ani')
                            summary_list[9] = all_fastani_dict.get(
                                userleaf.taxon.label).get(pplacer_leafnode).get('af')
                        summary_list[11] = 'topological placement and ANI have incongruent species assignments'
                        if len(sorted_dict) > 0:
                            summary_list[12] = '; '.join(self._formatnote(
                                sorted_dict, [fastani_matching_reference, pplacer_leafnode]))

                        summaryfout.write("{}\n".format(
                            '\t'.join(['N/A' if x is None else str(x) for x in summary_list])))

                    classified_user_genomes.append(userleaf.taxon.label)
            elif userleaf.taxon.label in all_fastani_dict:
                sorted_dict = sorted(all_fastani_dict.get(
                    userleaf.taxon.label).iteritems(), key=lambda(_x, y): y['ani'], reverse=True)
                fastani_matching_reference = sorted_dict[0][0]
                taxa_str = ";".join(self.gtdb_taxonomy.get(
                    add_ncbi_prefix(fastani_matching_reference))[:-1])
                summary_list[0] = userleaf.taxon.label
                summary_list[1] = self.standardise_taxonomy(taxa_str)
                summary_list[2] = fastani_matching_reference
                summary_list[3] = ";".join(self.gtdb_taxonomy.get(
                    add_ncbi_prefix(fastani_matching_reference)))
                summary_list[4] = all_fastani_dict.get(userleaf.taxon.label).get(
                    fastani_matching_reference).get('ani')
                summary_list[5] = all_fastani_dict.get(userleaf.taxon.label).get(
                    fastani_matching_reference).get('af')
                summary_list[10] = 'ANI/Placement'
                summary_list[11] = 'topological placement and ANI have incongruent species assignments'
                if len(sorted_dict) > 0:
                    summary_list[12] = '; '.join(self._formatnote(
                        sorted_dict, [fastani_matching_reference]))

                summaryfout.write("{}\n".format(
                    '\t'.join(['N/A' if x is None else str(x) for x in summary_list])))

                classified_user_genomes.append(userleaf.taxon.label)
        return classified_user_genomes

    def _fastaniWorker(self, sublist_genomes, genomes, out_q):
        """Multi thread worker to calculate FastANI"""
        try:
            for userleaf, potential_nodes in sublist_genomes.iteritems():
                dict_parser_distance = self._calculate_fastani_distance(
                    userleaf, potential_nodes, genomes)
                for k, v in dict_parser_distance.iteritems():
                    if k in out_q:
                        raise Exception("{} not in output.".format(k))
                    out_q[k] = v
            return True
        except Exception as error:
            print error
            raise

    def _parse_subtree(self, cur_node, dict_subrank, lower_rank):
        for childn in cur_node.child_nodes():
            if childn.is_leaf():
                continue
            elif childn.label is not None and childn.label.startswith(lower_rank):
                dict_subrank[childn.label] = abs(1 - childn.edge_length)
            else:
                dict_subrank = self._parse_subtree(
                    childn, dict_subrank, lower_rank)
        return dict_subrank

    def _get_redtax(self, list_subnode, closest_rank):
        """
        Provide a taxonomy string to a user genome based on the reference genomes of the same clade.
        If the clade contains multiple reference genomes we are comparing their taxonomies.
        -If all reference genomes have the same taxonomy up to the 'closest rank' ,
        the taxonomy string including the closest rank is returned.
        -If **NOT** all reference genomes have the same taxonomy up to the 'closest rank',
        the taxonomy string **NOT** including the closest rank is returned.

        Parameters
        ----------
        list_subnode : list of leaf nodes including multiple reference genome.
        closest_rank : last rank of the reference taxonomy

        Returns
        -------
        string
            Taxonomy string.

        """

        subtax, multirefrank = self._parse_subnodes(list_subnode, closest_rank)
        # if all orders in the list are the same, the user genomes gets the
        # same order
        if len(set(multirefrank)) == 1:
            # case d
            subtax.append(multirefrank[0])
        else:
            # otherwise it's stored as undefined
            # case a,b
            subtax.append(closest_rank + "undefined")
        return ';'.join(subtax)

    def _parse_subnodes(self, list_subnode, closest_rank):
        subtax = []
        multirefrank = []
        initial_loop = True
        for item in list_subnode:
            # We get the taxonomy of all reference genomes
            if item.startswith('RS_') or item.startswith('GB_') or item.startswith('UBA'):
                taxonomy_from_file = self.gtdb_taxonomy.get(item)
                # we store the selected rank (i.e. order) for each reference
                # genome
                for rank in taxonomy_from_file:
                    if rank.startswith(closest_rank):
                        multirefrank.append(rank)
                        initial_loop = False
                        break
                    elif initial_loop:
                        # The first iteration is used to stored upper level (
                        # i.e. domain,phylum,class )
                        subtax.append(rank)
        return subtax, multirefrank

    def _calculate_red_distances(self, input_tree, out_dir):
        """
        Provide a taxonomy string to a user genome based on the reference genomes of the same clade.
        If the clade contains multiple reference genomes we are comparing their taxonomies.
        -If all reference genomes have the same taxonomy up to the 'closest rank' ,
        the taxonomy string including the closest rank is returned.
        -If **NOT** all reference genomes have the same taxonomy up to the 'closest rank',
        the taxonomy string **NOT** including the closest rank is returned.

        Parameters
        ----------
        list_subnode : list of leaf nodes including multiple reference genome.
        closest_rank : last rank of the reference taxonomy

        Returns
        -------
        string
            Taxonomy string.
        """

        # read tree
        self.logger.info('Reading tree.')
        tree = dendropy.Tree.get_from_path(input_tree,
                                           schema='newick',
                                           rooting='force-rooted',
                                           preserve_underscores=True)

        self.logger.info('Reading taxonomy from file.')
        taxonomy = Taxonomy().read(Config.TAXONOMY_FILE)

        # determine taxa to be used for inferring distribution
        trusted_taxa = None
        taxa_for_dist_inference = self._filter_taxa_for_dist_inference(tree,
                                                                       taxonomy,
                                                                       trusted_taxa,
                                                                       Config.RED_MIN_CHILDREN,
                                                                       Config.RED_MIN_SUPPORT)

        phylum_rel_dists, rel_node_dists = self.median_rd_over_phyla(tree,
                                                                     taxa_for_dist_inference,
                                                                     taxonomy)

        # set edge lengths to median value over all rootings
        tree.seed_node.rel_dist = 0.0
        for n in tree.preorder_node_iter(lambda n: n != tree.seed_node):
            n.rel_dist = np_median(rel_node_dists[n.id])
            rd_to_parent = n.rel_dist - n.parent_node.rel_dist
            if rd_to_parent < 0:
                # This can occur since we are setting all nodes
                # to their median RED value.
                # self.logger.warning('Not all branches are positive after scaling.')
                pass
            n.edge_length = rd_to_parent

        if False:
            # These plots can be useful for debugging and internal use,
            # but are likely to be confusing to users.
            rd = RelativeDistance()

            input_tree_name = os.path.splitext(os.path.basename(input_tree))[0]
            plot_file = os.path.join(out_dir, '{}.png'.format(input_tree_name))
            rd._distribution_summary_plot(
                phylum_rel_dists, taxa_for_dist_inference, plot_file)

            gtdb_parent_ranks = Taxonomy().parents(taxonomy)
            median_outlier_table = os.path.join(
                out_dir, '{}.tsv'.format(input_tree_name))
            median_rank_file = os.path.join(
                out_dir, '{}.dict'.format(input_tree_name))
            rd._median_summary_outlier_file(phylum_rel_dists,
                                            taxa_for_dist_inference,
                                            gtdb_parent_ranks,
                                            median_outlier_table,
                                            median_rank_file,
                                            False)

            input_tree_name = os.path.splitext(os.path.basename(input_tree))[0]
            output_tree = os.path.join(
                out_dir, '{}.scaled.tree'.format(input_tree_name))
            tree.write_to_path(output_tree,
                               schema='newick',
                               suppress_rooting=True,
                               unquoted_underscores=True)

        return tree

    def _calculate_fastani_distance(self, user_leaf, list_leaf, genomes):
        """ Calculate the FastANI distance between all user genomes and the reference to classfy them at the species level

        Parameters
        ----------
        user_leaf : User genome node
        list_leaf : Dictionary of nodes including one or many user genomes and one reference genome.
        genomes : Dictionary of user genomes d[genome_id] -> FASTA file

        Returns
        -------
        dictionary
            dict_results[user_g]={ref_genome1:{"af":af,"ani":ani},ref_genome2:{"af":af,"ani":ani}}
        """
        try:
            self.tmp_output_dir = tempfile.mkdtemp()
            make_sure_path_exists(self.tmp_output_dir)

            # we write the two input files for fastani, the query file and
            # reference file
            query_list_file = open(os.path.join(
                self.tmp_output_dir, 'query_list.txt'), 'w')
            query_list_file.write('{0}\n'.format(
                genomes.get(user_leaf.taxon.label)))
            query_list_file.close()

            ref_list_file = open(os.path.join(
                self.tmp_output_dir, 'ref_list.txt'), 'w')
            leafnodes = list_leaf.get("potential_g")
            for node in leafnodes:
                leafnode = node[0]
                shortleaf = leafnode.taxon.label
                if leafnode.taxon.label.startswith('GB_') or leafnode.taxon.label.startswith('RS_'):
                    shortleaf = leafnode.taxon.label[3:]
                ref_list_file.write('{}\n'.format(os.path.join(
                    Config.FASTANI_GENOMES, shortleaf + Config.FASTANI_GENOMES_EXT)))
            ref_list_file.close()

            # run fastANI
            if not os.path.isfile(os.path.join(self.tmp_output_dir, 'query_list.txt')) or not os.path.isfile(os.path.join(self.tmp_output_dir, 'ref_list.txt')):
                raise

            cmd = 'fastANI --ql {0} --rl {1} -o {2} > /dev/null 2>{3}'.format(os.path.join(self.tmp_output_dir, 'query_list.txt'),
                                                                              os.path.join(
                                                                                  self.tmp_output_dir, 'ref_list.txt'),
                                                                              os.path.join(
                                                                                  self.tmp_output_dir, 'results.tab'),
                                                                              os.path.join(self.tmp_output_dir, 'error.log'))
            os.system(cmd)

            if not os.path.isfile(os.path.join(self.tmp_output_dir, 'results.tab')):
                errstr = 'FastANI has stopped:\n'
                if os.path.isfile(os.path.join(self.tmp_output_dir, 'error.log')):
                    with open(os.path.join(self.tmp_output_dir, 'error.log')) as debug:
                        for line in debug:
                            finalline = line
                        errstr += finalline
                raise ValueError(errstr)

            dict_parser_distance = self._parse_fastani_results(
                os.path.join(self.tmp_output_dir, 'results.tab'))
            shutil.rmtree(self.tmp_output_dir)
            return dict_parser_distance

        except ValueError as error:
            if os.path.exists(self.tmp_output_dir):
                shutil.rmtree(self.tmp_output_dir)
            raise error
        except Exception as error:
            if os.path.exists(self.tmp_output_dir):
                shutil.rmtree(self.tmp_output_dir)
            raise error

    def _parse_fastani_results(self, fastout_file):
        """ Parse the fastani output file


        Parameters
        ----------
        fastout_file : fastani output file.


        Returns
        -------
        dictionary
            dict_results[user_g]={ref_genome1:{"af":af,"ani":ani},ref_genome2:{"af":af,"ani":ani}}
        """
        dict_results = {}
        with open(fastout_file) as fastfile:
            for line in fastfile:
                info = line.strip().split()
                ref_genome = os.path.basename(info[1]).replace(
                    Config.FASTANI_GENOMES_EXT, "")
                user_g = remove_extension(os.path.basename(info[0]))
                ani = round(float(info[2]), 2)
                af = round(float(info[3]) / float(info[4]), 2)
                if user_g in dict_results and Config.FASTANI_SPECIES_THRESHOLD <= ani:
                    dict_results[user_g][ref_genome] = {"ani": ani, 'af': af}
                elif Config.FASTANI_SPECIES_THRESHOLD <= ani:
                    dict_results[user_g] = {ref_genome: {"ani": ani, "af": af}}

        return dict_results

    def _filter_taxa_for_dist_inference(self, tree, taxonomy, trusted_taxa, min_children, min_support):
        """Determine taxa to use for inferring distribution of relative divergences.

        Parameters
        ----------
        tree : Dendropy Tree
            Phylogenetic tree.
        taxonomy : d[taxon ID] -> [d__x; p__y; ...]
            Taxonomy for each taxon.
        trusted_taxa : iterable
            Trusted taxa to consider when inferring distribution.
        min_children : int
            Only consider taxa with at least the specified number of children taxa when inferring distribution.
        min_support : float
            Only consider taxa with at least this level of support when inferring distribution.
        """

        # determine children taxa for each named group
        taxon_children = Taxonomy().taxon_children(taxonomy)

        # get all named groups
        taxa_for_dist_inference = set()
        for taxon_id, taxa in taxonomy.iteritems():
            for taxon in taxa:
                taxa_for_dist_inference.add(taxon)

        # sanity check species names as these are a common problem
        species = set()
        for taxon_id, taxa in taxonomy.iteritems():
            if len(taxa) > Taxonomy.rank_index['s__']:
                species_name = taxa[Taxonomy.rank_index['s__']]
                valid, error_msg = True, None
                if species_name != 's__':
                    valid, error_msg = Taxonomy().validate_species_name(
                        species_name, require_full=True, require_prefix=True)
                if not valid:
                    print '[Warning] Species name {} for {} is invalid: {}'.format(species_name, taxon_id, error_msg)
                    continue

                species.add(species_name)

        # restrict taxa to those with a sufficient number of named children
        # Note: a taxonomic group with no children will not end up in the
        # taxon_children data structure so care must be taken when applying
        # this filtering criteria.
        if min_children > 0:
            valid_taxa = set()
            for taxon, children_taxa in taxon_children.iteritems():
                if len(children_taxa) >= min_children:
                    valid_taxa.add(taxon)

            taxa_for_dist_inference.intersection_update(valid_taxa)

            # explicitly add in the species since they have no
            # children and thus be absent from the taxon_child dictionary
            taxa_for_dist_inference.update(species)

        # restrict taxa used for inferring distribution to those with
        # sufficient support
        if min_support > 0:
            for node in tree.preorder_node_iter():
                if not node.label or node.is_leaf():
                    continue

                # check for support value
                support, taxon_name, _auxiliary_info = parse_label(node.label)

                if not taxon_name:
                    continue

                if support and float(support) < min_support:
                    taxa_for_dist_inference.difference_update([taxon_name])
                elif not support and min_support > 0:
                    # no support value, so inform user if they were trying to
                    # filter on this property
                    print '[Error] Tree does not contain support values. As such, --min_support should be set to 0.'
                    continue

        # restrict taxa used for inferring distribution to the trusted set
        if trusted_taxa:
            taxa_for_dist_inference = trusted_taxa.intersection(
                taxa_for_dist_inference)

        return taxa_for_dist_inference

    def median_rd_over_phyla(self,
                             tree,
                             taxa_for_dist_inference,
                             taxonomy):
        """Calculate the median relative divergence over all phyla rootings.

        Parameters
        ----------
        tree : Tree
          Dendropy tree.
        taxa_for_dist_inference : set
          Taxa to use for inference relative divergence distributions.
        taxonomy : d[taxon_id] -> [d__, p__, ..., s__]
          Taxonomy of extant taxa.
        """

        # get list of phyla level lineages
        all_phyla = self._get_phyla_lineages(tree)
        self.logger.info('Identified %d phyla.' % len(all_phyla))

        phyla = [p for p in all_phyla if p in taxa_for_dist_inference]
        self.logger.info(
            'Using %d phyla as rootings for inferring RED distributions.' % len(phyla))
        if len(phyla) < 2:
            self.logger.error('Rescaling requires at least 2 valid phyla.')
            sys.exit(-1)

        # give each node a unique id
        for i, n in enumerate(tree.preorder_node_iter()):
            n.id = i

        # calculate relative divergence for tree rooted on each phylum
        phylum_rel_dists = {}
        rel_node_dists = defaultdict(list)
        rd = RelativeDistance()
        for p in phyla:
            phylum = p.replace('p__', '').replace(' ', '_').lower()
            status_msg = '==> Calculating information with rooting on {}.              '.format(
                phylum.capitalize())
            sys.stdout.write('{}\r'.format(status_msg))
            sys.stdout.flush()

            cur_tree = self.root_with_outgroup(tree, taxonomy, p)

            # calculate relative distance to taxa
            rel_dists = rd.rel_dist_to_named_clades(cur_tree)
            rel_dists.pop(0, None)  # remove results for Domain

            # remove named groups in outgroup
            children = Taxonomy().children(p, taxonomy)
            for r in rel_dists.keys():
                rel_dists[r].pop(p, None)

            for t in children:
                for r in rel_dists.keys():
                    rel_dists[r].pop(t, None)

            phylum_rel_dists[phylum] = rel_dists

            # calculate relative distance to all nodes
            rd.decorate_rel_dist(cur_tree)

            # determine which lineages represents the 'ingroup'
            ingroup_subtree = None
            for c in cur_tree.seed_node.child_node_iter():
                _support, taxon_name, _auxiliary_info = parse_label(c.label)
                if not taxon_name or p not in taxon_name:
                    ingroup_subtree = c
                    break

            # do a preorder traversal of 'ingroup' and record relative
            # divergence to nodes
            for n in ingroup_subtree.preorder_iter():
                rel_node_dists[n.id].append(n.rel_dist)

        sys.stdout.write(
            '==> Inference for RED distributions finished.                         ')
        sys.stdout.flush()
        sys.stdout.write('\n')

        return phylum_rel_dists, rel_node_dists

    def _get_phyla_lineages(self, tree):
        """Get list of phyla level lineages.

        Parameters
        ----------
        tree : Dendropy Tree
            Phylogenetic tree.

        Returns
        -------
        list
            List of phyla level lineages.
        """
        phyla = []
        for node in tree.preorder_node_iter():
            if not node.label or node.is_leaf():
                continue

            _support, taxon_name, _auxiliary_info = parse_label(node.label)
            if taxon_name:
                taxa = [x.strip() for x in taxon_name.split(';')]
                if taxa[-1].startswith('p__'):
                    phyla.append(taxa[-1])

        return phyla

    def root_with_outgroup(self, input_tree, taxonomy, outgroup_taxa):
        """Reroot the tree using the given outgroup.

        Parameters
        ----------
        input_tree : Dendropy Tree
          Tree to rerooted.
        taxonomy : dict
            Taxonomy for taxa.
        outgroup : iterable
          Labels of taxa in outgroup.

        Returns
        -------
        Dendropy Tree
            Deep-copy of original tree rerooted on outgroup.
        """

        new_tree = input_tree.clone()

        outgroup = set()
        for genome_id, taxa in taxonomy.iteritems():
            if outgroup_taxa in taxa:
                outgroup.add(genome_id)

        outgroup_in_tree = set()
        ingroup_in_tree = set()
        for n in new_tree.leaf_node_iter():
            if n.taxon.label in outgroup:
                outgroup_in_tree.add(n.taxon)
            else:
                ingroup_in_tree.add(n)

        if len(outgroup_in_tree) == 0:
            self.logger.warning('No outgroup taxa identified in the tree.')
            self.logger.warning('Tree was not rerooted.')
            sys.exit(0)

        # There is a complication here. We wish to find the MRCA of the outgroup
        # taxa. Finding the MRCA requires a rooted tree and we have no gaurantee
        # that the tree isn't currently rooted within the outgroup clade. There is
        # also no way to identify a node that is gauranteed to be outside the outgroup
        # clade. As such, the tree is randomly rooted on a leaf node not in the outgroup.
        # This random rerooting is performed until the MRCA does not spans all taxa in
        # the tree.

        leaves_in_tree = sum([1 for _ in new_tree.leaf_node_iter()])
        while True:
            rnd_ingroup_leaf = random.sample(ingroup_in_tree, 1)[0]
            new_tree.reroot_at_edge(rnd_ingroup_leaf.edge,
                                    length1=0.5 * rnd_ingroup_leaf.edge_length,
                                    length2=0.5 * rnd_ingroup_leaf.edge_length)

            mrca = new_tree.mrca(taxa=outgroup_in_tree)
            leaves_in_mrca = sum([1 for _ in mrca.leaf_iter()])
            if leaves_in_mrca != leaves_in_tree:
                break

        if leaves_in_mrca == leaves_in_tree:
            self.logger.error('The MRCA spans all taxa in the tree.')
            self.logger.error(
                'This indicating the selected outgroup is likely polyphyletic in the current tree.')
            self.logger.error(
                'This should never occur. Please report this as a bug.')
            sys.exit(-1)

        if mrca.edge_length is None:
            # self.logger.info('Tree appears to already be rooted on this outgroup.')
            pass
        else:
            new_tree.reroot_at_edge(mrca.edge,
                                    length1=0.5 * mrca.edge_length,
                                    length2=0.5 * mrca.edge_length)

        return new_tree



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging
from collections import defaultdict

from biolib.seq_tk import trim_seqs

from biolib.common import remove_extension
from biolib.seq_io import read_fasta
from biolib.taxonomy import Taxonomy
from biolib.external.execute import check_dependencies

from external.prodigal import Prodigal
from external.tigrfam_search import TigrfamSearch
from external.pfam_search import PfamSearch
from external.hmm_aligner import HmmAligner

import config.config as Config
import config.config_metadata as ConfigMetadata

from tools import merge_two_dicts


class Markers(object):
    """Identify and align marker genes."""

    def __init__(self, cpus=1):
        """Initialize."""

        self.logger = logging.getLogger('timestamp')

        self.cpus = cpus

        self.genome_file_suffix = ConfigMetadata.GENOME_FILE_SUFFIX
        self.protein_file_suffix = ConfigMetadata.PROTEIN_FILE_SUFFIX
        self.nt_gene_file_suffix = ConfigMetadata.NT_GENE_FILE_SUFFIX
        self.gff_file_suffix = ConfigMetadata.GFF_FILE_SUFFIX
        self.checksum_suffix = ConfigMetadata.CHECKSUM_SUFFIX

        self.taxonomy_file = Config.TAXONOMY_FILE

        self.pfam_hmm_dir = Config.PFAM_HMM_DIR
        self.pfam_suffix = ConfigMetadata.PFAM_SUFFIX
        self.pfam_top_hit_suffix = ConfigMetadata.PFAM_TOP_HIT_SUFFIX

        self.tigrfam_hmms = Config.TIGRFAM_HMMS
        self.tigrfam_suffix = ConfigMetadata.TIGRFAM_SUFFIX
        self.tigrfam_top_hit_suffix = ConfigMetadata.TIGRFAM_TOP_HIT_SUFFIX

    def _report_identified_marker_genes(self, gene_dict, outdir, prefix):
        """Report statistics for identified marker genes."""

        bac_outfile = open(os.path.join(
            outdir, prefix + "_bac120_markers_summary.tsv"), "w")
        arc_outfile = open(os.path.join(
            outdir, prefix + "_ar122_markers_summary.tsv"), "w")

        header = "Name\tnumber_unique_genes\tnumber_multiple_genes\tnumber_missing_genes\tlist_unique_genes\tlist_multiple_genes\tlist_missing_genes\n"

        bac_outfile.write(header)
        arc_outfile.write(header)

        # gather information for all marker genes
        marker_dbs = {"PFAM": ConfigMetadata.PFAM_TOP_HIT_SUFFIX,
                      "TIGR": ConfigMetadata.TIGRFAM_TOP_HIT_SUFFIX}

        marker_bac_list_original = []
        for db_marker in Config.BAC120_MARKERS.keys():
            marker_bac_list_original.extend([marker.replace(".HMM", "").replace(".hmm", "")
                                             for marker in Config.BAC120_MARKERS[db_marker]])

        marker_arc_list_original = []
        for db_marker in Config.AR122_MARKERS.keys():
            marker_arc_list_original.extend([marker.replace(".HMM", "").replace(".hmm", "")
                                             for marker in Config.AR122_MARKERS[db_marker]])

        for db_genome_id, info in gene_dict.items():
            unique_genes_bac, multi_hits_bac, missing_genes_bac = [], [], []
            unique_genes_arc, multi_hits_arc, missing_genes_arc = [], [], []

            gene_bac_dict, gene_arc_dict = {}, {}

            path = info.get("aa_gene_path")
            for _marker_db, marker_suffix in marker_dbs.iteritems():
                # get all gene sequences
                protein_file = str(path)
                tophit_path = protein_file.replace(
                    ConfigMetadata.PROTEIN_FILE_SUFFIX, marker_suffix)

                # we load the list of all the genes detected in the genome
                all_genes_dict = read_fasta(protein_file, False)

                # Prodigal adds an asterisks at the end of each called genes.
                # These asterisks sometimes appear in the MSA, which can be
                # an issue for some downstream software
                for seq_id, seq in all_genes_dict.iteritems():
                    if seq[-1] == '*':
                        all_genes_dict[seq_id] = seq[:-1]

                # we store the tophit file line by line and store the
                # information in a dictionary
                with open(tophit_path) as tp:
                    # first line is header line
                    tp.readline()

                    for line_tp in tp:
                        linelist = line_tp.split("\t")
                        genename = linelist[0]
                        sublist = linelist[1]
                        if ";" in sublist:
                            diff_markers = sublist.split(";")
                        else:
                            diff_markers = [sublist]

                        for each_mark in diff_markers:
                            sublist = each_mark.split(",")
                            markerid = sublist[0]

                            if (markerid not in marker_bac_list_original and
                                    markerid not in marker_arc_list_original):
                                continue

                            if markerid in marker_bac_list_original:
                                if markerid in gene_bac_dict:
                                    gene_bac_dict.get(markerid)[
                                        "multihit"] = True
                                else:
                                    gene_bac_dict[markerid] = {
                                        "gene": genename,
                                        "multihit": False}

                            if markerid in marker_arc_list_original:
                                if markerid in gene_arc_dict:
                                    gene_arc_dict.get(markerid)[
                                        "multihit"] = True
                                else:
                                    gene_arc_dict[markerid] = {
                                        "gene": genename,
                                        "multihit": False}

            for mid in marker_bac_list_original:
                if mid not in gene_bac_dict:
                    missing_genes_bac.append(mid)
                elif gene_bac_dict[mid]["multihit"]:
                    multi_hits_bac.append(mid)
                else:
                    unique_genes_bac.append(mid)

            for mid in marker_arc_list_original:
                if mid not in gene_arc_dict:
                    missing_genes_arc.append(mid)
                elif gene_arc_dict[mid]["multihit"]:
                    multi_hits_arc.append(mid)
                else:
                    unique_genes_arc.append(mid)

            bac_outfile.write("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\n".format(db_genome_id,
                                                                           len(unique_genes_bac),
                                                                           len(multi_hits_bac),
                                                                           len(missing_genes_bac),
                                                                           ','.join(
                                                                               unique_genes_bac),
                                                                           ','.join(
                                                                               multi_hits_bac),
                                                                           ','.join(missing_genes_bac)))

            arc_outfile.write("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\n".format(db_genome_id,
                                                                           len(unique_genes_arc),
                                                                           len(multi_hits_arc),
                                                                           len(missing_genes_arc),
                                                                           ','.join(
                                                                               unique_genes_arc),
                                                                           ','.join(
                                                                               multi_hits_arc),
                                                                           ','.join(missing_genes_arc)))

        bac_outfile.close()
        arc_outfile.close()

    def identify(self,
                 genomes,
                 out_dir,
                 prefix):
        """Identify marker genes in genomes."""

        check_dependencies(['prodigal', 'hmmsearch'])

        try:
            self.logger.info('Identifying markers in %d genomes with %d threads.' % (len(genomes),
                                                                                     self.cpus))

            self.logger.info("Running Prodigal to identify genes.")
            self.marker_gene_dir = os.path.join(
                out_dir, Config.MARKER_GENE_DIR)
            prodigal = Prodigal(self.cpus,
                                False,
                                self.marker_gene_dir,
                                self.protein_file_suffix,
                                self.nt_gene_file_suffix,
                                self.gff_file_suffix)
            genome_dictionary = prodigal.run(genomes)

            # annotated genes against TIGRFAM and Pfam databases
            self.logger.info("Identifying TIGRFAM protein families.")
            gene_files = [genome_dictionary[db_genome_id]['aa_gene_path']
                          for db_genome_id in genome_dictionary.keys()]

            tigr_search = TigrfamSearch(self.cpus,
                                        self.tigrfam_hmms,
                                        self.protein_file_suffix,
                                        self.tigrfam_suffix,
                                        self.tigrfam_top_hit_suffix,
                                        self.checksum_suffix,
                                        self.marker_gene_dir)
            tigr_search.run(gene_files)

            self.logger.info("Identifying Pfam protein families.")
            pfam_search = PfamSearch(self.cpus,
                                     self.pfam_hmm_dir,
                                     self.protein_file_suffix,
                                     self.pfam_suffix,
                                     self.pfam_top_hit_suffix,
                                     self.checksum_suffix,
                                     self.marker_gene_dir)
            pfam_search.run(gene_files)

            self._report_identified_marker_genes(
                genome_dictionary, out_dir, prefix)

        except IOError as e:
            self.logger.error(str(e))
            self.logger.error("GTDB-Tk has encountered an error.")

        except Exception as e:
            self.logger.error(str(e))
            raise

    def _path_to_identify_data(self, identity_dir):
        """Get path to genome data produced by 'identify' command."""

        marker_gene_dir = os.path.join(identity_dir, Config.MARKER_GENE_DIR)

        genomic_files = {}
        for gid in os.listdir(marker_gene_dir):
            gid_dir = os.path.join(marker_gene_dir, gid)
            if not os.path.isdir(gid_dir):
                continue

            genomic_files[gid] = {'aa_gene_path': os.path.join(gid_dir, gid + self.protein_file_suffix),
                                  'translation_table_path': os.path.join(gid_dir, 'prodigal_translation_table.tsv'),
                                  'nt_gene_path': os.path.join(gid_dir, gid + self.nt_gene_file_suffix),
                                  'gff_path': os.path.join(gid_dir, gid + self.gff_file_suffix)
                                  }
        return genomic_files

    def _msa_filter_by_taxa(self, concatenated_file, gtdb_taxonomy, taxa_filter, outgroup_taxon):
        """Filter GTDB MSA filtered to specified taxa."""

        msa = read_fasta(concatenated_file)
        self.logger.info(
            'Read concatenated alignment for %d GTDB genomes.' % len(msa))

        if taxa_filter is not None:
            taxa_to_keep = set(taxa_filter.split(','))

            if outgroup_taxon not in taxa_to_keep and outgroup_taxon is not None:
                taxa_to_keep.add(outgroup_taxon)

            filtered_genomes = 0
            for genome_id, taxa in gtdb_taxonomy.iteritems():
                common_taxa = taxa_to_keep.intersection(taxa)
                if len(common_taxa) == 0:
                    if genome_id in msa:
                        del msa[genome_id]
                        filtered_genomes += 1

            self.logger.info(
                'Filtered %d taxa based on assigned taxonomy.' % filtered_genomes)

        return msa

    def _apply_mask(self, gtdb_msa, user_msa, msa_mask, min_perc_aa):
        """Apply canonical mask to MSA file."""

        aligned_genomes = merge_two_dicts(gtdb_msa, user_msa)

        mask = open(msa_mask).readline().strip()

        if len(mask) != len(aligned_genomes.values()[0]):
            self.logger.error('Mask and alignment length do not match.')
            sys.exit()

        output_seqs = {}
        pruned_seqs = {}
        for seq_id, seq in aligned_genomes.iteritems():
            masked_seq = ''.join(
                [seq[i] for i in xrange(0, len(mask)) if mask[i] == '1'])

            valid_bases = len(masked_seq) - \
                masked_seq.count('.') - masked_seq.count('-')
            if seq_id in user_msa and valid_bases < len(masked_seq) * min_perc_aa:
                pruned_seqs[seq_id] = masked_seq
                continue

            output_seqs[seq_id] = masked_seq

        return output_seqs, pruned_seqs

    def _write_msa(self, seqs, output_file, gtdb_taxonomy):
        """Write sequences to FASTA file."""

        fout = open(output_file, 'w')
        for genome_id, alignment in seqs.iteritems():
            if genome_id in gtdb_taxonomy:
                fout.write('>%s %s\n' %
                           (genome_id, ';'.join(gtdb_taxonomy[genome_id])))
            else:
                fout.write('>%s\n' % genome_id)
            fout.write('%s\n' % alignment)
        fout.close()

    def _genome_domain(self, identity_dir, prefix):
        """Determine domain of User genomes based on identified marker genes."""

        bac_count = defaultdict(int)
        ar_count = defaultdict(int)
        for d, marker_file in ((bac_count, prefix + '_bac120_markers_summary.tsv'),
                               (ar_count, prefix + '_ar122_markers_summary.tsv')):
            with open(os.path.join(identity_dir, marker_file)) as f:
                f.readline()

                for line in f:
                    line_split = line.strip().split('\t')

                    gid = line_split[0]
                    num_markers = int(line_split[1])

                    d[gid] = num_markers

        bac_gids = set()
        ar_gids = set()
        for gid in bac_count:
            arc_aa_per = (ar_count[gid] * 100.0 / Config.AR_MARKER_COUNT)
            bac_aa_per = (bac_count[gid] * 100.0 / Config.BAC_MARKER_COUNT)
            if bac_aa_per >= arc_aa_per:
                bac_gids.add(gid)
            else:
                ar_gids.add(gid)

        return bac_gids, ar_gids

    def align(self,
              identify_dir,
              taxa_filter,
              min_perc_aa,
              custom_msa_filters,
              consensus,
              min_per_taxa,
              out_dir,
              prefix,
              outgroup_taxon):
        """Align marker genes in genomes."""

        try:
            genomic_files = self._path_to_identify_data(identify_dir)
            self.logger.info('Aligning markers in %d genomes with %d threads.' % (len(genomic_files),
                                                                                  self.cpus))

            # determine marker set for each user genome
            bac_gids, ar_gids = self._genome_domain(identify_dir, prefix)

            # align user genomes
            gtdb_taxonomy = Taxonomy().read(self.taxonomy_file)
            for gids, msa_file, mask_file, marker_set_id in ((bac_gids, Config.CONCAT_BAC120, Config.MASK_BAC120, "bac120"),
                                                             (ar_gids, Config.CONCAT_AR122, Config.MASK_AR122, "ar122")):

                if len(gids) == 0:
                    continue

                if marker_set_id == 'bac120':
                    self.logger.info(
                        'Processing %d genomes identified as bacterial.' % len(gids))
                else:
                    self.logger.info(
                        'Processing %d genomes identified as archaeal.' % len(gids))

                cur_genome_files = {
                    gid: f for gid, f in genomic_files.iteritems() if gid in gids}

                gtdb_msa = self._msa_filter_by_taxa(msa_file,
                                                    gtdb_taxonomy,
                                                    taxa_filter,
                                                    outgroup_taxon)
                gtdb_msa_mask = os.path.join(Config.MASK_DIR, mask_file)

                hmm_aligner = HmmAligner(self.cpus,
                                         self.pfam_top_hit_suffix,
                                         self.tigrfam_top_hit_suffix,
                                         self.protein_file_suffix,
                                         self.pfam_hmm_dir,
                                         self.tigrfam_hmms,
                                         Config.BAC120_MARKERS,
                                         Config.AR122_MARKERS,
                                         Config.RPS23_MARKERS)
                user_msa = hmm_aligner.align_marker_set(cur_genome_files,
                                                        marker_set_id)

                # filter columns without sufficient representation across taxa
                if custom_msa_filters:
                    aligned_genomes = merge_two_dicts(gtdb_msa, user_msa)
                    self.logger.info(
                        'Trimming columns with insufficient taxa or poor consensus.')
                    trimmed_seqs, pruned_seqs, count_wrong_pa, count_wrong_cons = trim_seqs(aligned_genomes,
                                                                                            min_per_taxa / 100.0,
                                                                                            consensus / 100.0,
                                                                                            min_perc_aa / 100.0)
                    self.logger.info(('Trimmed alignment from %d to %d AA (%d by minimum taxa percent, ' +
                                      '%d by consensus).') % (len(aligned_genomes.values()[0]),
                                                              len(trimmed_seqs.values()[
                                                                  0]),
                                                              count_wrong_pa,
                                                              count_wrong_cons))

                    self.logger.info('%d taxa have amino acids in <%.1f%% of columns in filtered MSA.' % (
                        len(pruned_seqs),
                        min_perc_aa))

                    pruned_user_genomes = set(
                        pruned_seqs).intersection(user_msa)
                    if len(pruned_user_genomes):
                        self.logger.info(
                            'Pruned genomes include %d user submitted genomes.' % len(pruned_user_genomes))
                else:
                    self.logger.info(
                        'Masking columns of multiple sequence alignment.')
                    trimmed_seqs, pruned_seqs = self._apply_mask(gtdb_msa,
                                                                 user_msa,
                                                                 gtdb_msa_mask,
                                                                 min_perc_aa / 100.0)
                    self.logger.info('Masked alignment from %d to %d AA.' % (len(gtdb_msa.values()[0]),
                                                                             len(trimmed_seqs.values()[0])))

                    if min_perc_aa > 0:
                        self.logger.info('%d user genomes have amino acids in <%.1f%% of columns in filtered MSA.' % (
                            len(pruned_seqs),
                            min_perc_aa))

                # write out filtering information
                fout = open(os.path.join(out_dir, prefix +
                                         ".%s.filtered.tsv" % marker_set_id), 'w')
                for pruned_seq_id, pruned_seq in pruned_seqs.items():
                    valid_bases = len(
                        pruned_seq) - pruned_seq.count('.') - pruned_seq.count('-')
                    perc_alignment = valid_bases * 100.0 / len(pruned_seq)
                    fout.write('%s\t%s\n' % (pruned_seq_id,
                                             'Insufficient number of amino acids in MSA (%.1f%%)' % perc_alignment))
                fout.close()

                # write out MSA
                self.logger.info(
                    'Creating concatenated alignment for %d taxa.' % len(trimmed_seqs))
                msa_file = os.path.join(
                    out_dir, prefix + ".%s.msa.fasta" % marker_set_id)
                self._write_msa(trimmed_seqs, msa_file, gtdb_taxonomy)

                user_msa_file = os.path.join(
                    out_dir, prefix + ".%s.user_msa.fasta" % marker_set_id)
                trimmed_user_msa = {
                    k: v for k, v in trimmed_seqs.iteritems() if k in user_msa}
                self._write_msa(trimmed_user_msa, user_msa_file, gtdb_taxonomy)

                #==============================================================
                # #all_user_msa_file = os.path.join(out_dir, prefix + ".%s.user_msa.fasta" % marker_set_id)
                # trimmed_all_user_msa = {k:v for k, v in trimmed_seqs.iteritems() if k in user_msa}
                # pruned_all_user_msa = {k:v for k, v in pruned_seqs.iteritems() if k in user_msa}
                # all_user_msa = merge_two_dicts(trimmed_all_user_msa,pruned_all_user_msa)
                # self._write_msa(all_user_msa, all_user_msa_file, gtdb_taxonomy)
                #==============================================================

        except IOError as e:
            self.logger.error(str(e))
            self.logger.error("GTDB-Tk has encountered an error.")



#!/usr/bin/env python

from setuptools import setup

import os


def version():
    setupDir = os.path.dirname(os.path.realpath(__file__))
    versionFile = open(os.path.join(setupDir, 'gtdbtk', 'VERSION'))
    return versionFile.readline().strip()

setup(
    name='gtdbtk',
    version=version(),
    author='Pierre-Alain Chaumeil and Donovan Parks',
    author_email='p.chaumeil@uq.edu.au',
    maintainer='Pierre-Alain Chaumeil and Donovan Parks',
    maintainer_email='donovan.parks@gmail.com',
    packages=['gtdbtk','gtdbtk.config','gtdbtk.external','gtdbtk.external.Bio.Pfam.Active_site','gtdbtk.external.Bio.Pfam.HMM','gtdbtk.external.Bio.Pfam.Scan'],
    scripts=['bin/gtdbtk'],
    package_data={'gtdbtk': ['VERSION','MANIFEST.in'],
                  'gtdbtk.external': ['pfam_search.pl'],
                  'gtdbtk.external.Bio.Pfam.Active_site':['as_search.pm'],
                  'gtdbtk.external.Bio.Pfam.HMM':['HMM.pm','HMMIO.pm','HMMMatch.pm','HMMResults.pm','HMMResultsIO.pm','HMMSequence.pm','HMMUnit.pm'],
                  'gtdbtk.external.Bio.Pfam.Scan':['PfamScan.pm','Seq.pm'],
                  },
    
    url='http://pypi.python.org/pypi/gtdbtk/',
    license='GPL3',
    description='A toolkit for assigning objective taxonomic classifications to bacterial and archaeal genomes.',
    classifiers=[
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Natural Language :: English',
          'Programming Language :: Python :: 2.7',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          ],
    install_requires=["biolib >= 0.0.43",
                        "jinja2>=2.7.3",
                        "mpld3>=0.2",
                        "dendropy>=4.1.0"],
)
